from typing import List, Optional, Tuple

import typer
from openapi_schema_pydantic import Components, Reference, Schema

from openapi_python_generator.language_converters.python.jinja_config import JINJA_ENV, MODELS_TEMPLATE, ENUM_TEMPLATE
from openapi_python_generator.models import Model, Property

def type_converter(schema: Schema, required: bool = False) -> str:
    """
    Converts an OpenAPI type to a Python type.
    :param schema: Schema containing the type to be converted
    :param required: Flag indicating if the type is required by the class
    :return: The converted type
    """
    if required:
        pre_type = ""
        post_type = ""
    else:
        pre_type = "Optional["
        post_type = "]"

    if schema.allOf is not None:
        return pre_type + "Tuple[" + ','.join(type_converter(i, True) for i in schema.allOf) + "]" + post_type
    elif schema.oneOf is not None or schema.anyOf is not None:
        used = schema.oneOf if schema.oneOf is not None else schema.anyOf
        return pre_type + "Union[" + ','.join(type_converter(i, True) for i in used) + "]" + post_type
    elif schema.type == "string":
        return pre_type + "str" + post_type
    elif schema.type == "integer":
        return pre_type + "int" + post_type
    elif schema.type == "number":
        return pre_type + "float" + post_type
    elif schema.type == "boolean":
        return pre_type + "bool" + post_type
    elif schema.type == "array":
        retVal = pre_type + "List["
        if isinstance(schema.items, Reference):
            retVal += schema.items.ref.split("/")[-1]
        elif isinstance(schema.items, Schema):
            retVal += type_converter(schema.items, True)
        else:
            retVal += "Any"

        return retVal + "]" + post_type
    elif schema.type == "object":
        return pre_type + "Dict[str, Any]" + post_type
    elif schema.type is None or schema.type == "null":
        return pre_type + "Any" + post_type
    else:
        raise TypeError(f"Unknown type: {schema.type}")


def _generate_property_from_schema(name: str, schema: Schema, parentSchema: Optional[Schema] = None) -> Property:
    """
    Generates a property from a schema. It takes the type of the schema and converts it to a python type, and then
    creates the according property.
    :param name: Name of the schema
    :param schema: schema to be converted
    :param parentSchema: Component this belongs to
    :return: Property
    """
    required = parentSchema is not None and parentSchema.required is not None and name in parentSchema.required
    return Property(
        name=name,
        type=type_converter(schema, required),
        required=required,
        default=None if required else "None")


def _generate_property_from_reference(name: str, reference: Reference,
                                      parentSchema: Optional[Schema] = None) -> Tuple[
    Property, str]:
    """
    Generates a property from a reference. It takes the name of the reference as the type, and then
    returns a property type
    :param name: Name of the schema
    :param reference: reference to be converted
    :param parentSchema: Component this belongs to
    :return: Property and model to be imported by the file
    """
    required = parentSchema.required is not None and name in parentSchema.required
    import_model = reference.ref.split("/")[-1]
    return Property(
        name=name,
        type=import_model if required else f"Optional[{import_model}]",
        required=required,
        default=None if required else "None"), import_model


def generate_models(components: Components) -> List[Model]:
    """
    Receives components from an OpenAPI 3.0 specification and generates the models from it.
    It does so, by iterating over the components.schemas dictionary. For each schema, it checks if
    it is a normal schema (i.e. simple type like string, integer, etc.), a reference to another schema, or
    an array of types/references. It then computes pydantic models from it using jinja2
    :param components: The components from an OpenAPI 3.0 specification.
    :return: A list of models.
    """
    models = []

    for name, schema in components.schemas.items():
        if schema.enum is not None:
            m = Model(
                file_name=name,
                content=JINJA_ENV.get_template(ENUM_TEMPLATE).render(name=name,**schema.dict()),
                openapi_object=schema,
                references=[],
                properties=[],
            )
            try:
                compile(m.content, "<string>", "exec")
                models.append(m)
            except SyntaxError as e: #pragma: no cover
                typer.echo(f"Error in model {name}: {e}")

            continue  #pragma: no cover

        import_models = []
        properties = []
        for prop_name, property in schema.properties.items():
            if isinstance(property, Reference):
                conv_property, import_model = _generate_property_from_reference(prop_name, property, schema)
                import_models.append(import_model)
            else:
                conv_property = _generate_property_from_schema(prop_name, property, schema)
            properties.append(conv_property)

        generated_content = JINJA_ENV.get_template(MODELS_TEMPLATE).render(
            schema_name=name,
            schema=schema,
            properties=properties,
            import_models=import_models,
        )

        try:
            compile(generated_content, "<string>", "exec")
        except SyntaxError as e: #pragma: no cover
            typer.echo(f"Error in model {name}: {e}") #pragma: no cover

        models.append(Model(
            file_name=name,
            content=generated_content,
            openapi_object=schema,
            references=import_models,
            properties=properties,
        ))

    return models



from pathlib import Path
from typing import Optional, Union

import httpx
from httpx import ConnectError, ConnectTimeout
from orjson import orjson, JSONDecodeError
import typer
from pydantic import ValidationError
import autopep8

from openapi_schema_pydantic import OpenAPI
from .common import HTTPLibrary
from .language_converters.python.generator import generator
from .language_converters.python.jinja_config import JINJA_ENV, SERVICE_TEMPLATE
from .models import ConversionResult


def write_code(path: Path, content):
    """
    Write the content to the file at the given path.
    :param path: The path to the file.
    :param content: The content to write.
    """
    with open(path, 'w') as f:
        f.write(autopep8.fix_code(content))


def get_open_api(path: Union[str,Path]) -> OpenAPI:
    """
    Tries to fetch the openapi.json file from the web or load from a local file. Returns the according OpenAPI object.
    :param path:
    :return:
    """
    try:
        if not isinstance(path,Path) and (path.startswith("http://") or path.startswith("https://")):
            return OpenAPI(**orjson.loads(httpx.get(path).text))

        with open(path, 'r') as f:
            return OpenAPI(**orjson.loads(f.read()))
    except FileNotFoundError:
        typer.echo(f"File {path} not found. Please make sure to pass the path to the OpenAPI 3.0 specification.")
        raise
    except (ConnectError,ConnectTimeout):
        typer.echo(f"Could not connect to {path}.")
        raise ConnectError(f"Could not connect to {path}.")
    except (ValidationError,JSONDecodeError):
        typer.echo(f"File {path} is not a valid OpenAPI 3.0 specification, or there may be a problem with your JSON.")
        raise


def write_data(data: ConversionResult, output: str):
    """
    This function will firstly create the folderstrucutre of output, if it doesn't exist. Then it will create the
    models from data.models into the models sub module of the output folder. After this, the services will be created
    into the services sub module of the output folder.
    :param data: The data to write.
    :param output: The path to the output folder.
    """

    # Create the folder structure of the output folder.
    Path(output).mkdir(parents=True, exist_ok=True)

    # Create the models module.
    models_path = Path(output) / "models"
    models_path.mkdir(parents=True, exist_ok=True)

    # Create the services module.
    services_path = Path(output) / "services"
    services_path.mkdir(parents=True, exist_ok=True)

    files = []

    # Write the models.
    for model in data.models:
        files.append(model.file_name)
        write_code(models_path / f"{model.file_name}.py", model.content)

    # Create models.__init__.py file containing imports to all models.
    write_code(models_path / "__init__.py","\n".join([f'from {file} import *' for file in files]))

    files = []

    # Write the services.
    for service in data.services:
        files.append(service.file_name)
        write_code(services_path / f"{service.file_name}.py", JINJA_ENV.get_template(SERVICE_TEMPLATE).render(**service.dict()))

    # Create services.__init__.py file containing imports to all services.
    write_code(services_path / "__init__.py", "\n".join([f'from {file} import *' for file in files]))

    # Write the api_config.py file.
    write_code(Path(output) / "api_config.py", data.api_config.content)

    # Write the __init__.py file.
    write_code(Path(output) / "__init__.py", "from models import *\nfrom services import *\nfrom api_config import *")


def generate_data(file_name: str, output: str, library: Optional[HTTPLibrary] = HTTPLibrary.httpx) -> None:
    """
    Generate Python code from an OpenAPI 3.0 specification.
    """
    data = get_open_api(file_name)
    typer.echo(f"Generating data from {file_name}")
    result = generator(data)
    write_data(result, output)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['openapi_python_generator',
 'openapi_python_generator.language_converters',
 'openapi_python_generator.language_converters.python']

package_data = \
{'': ['*'],
 'openapi_python_generator.language_converters.python': ['templates/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'autopep8>=1.6.0,<2.0.0',
 'httpx[all]>=0.23.0,<0.24.0',
 'openapi-schema-pydantic>=1.2.3,<2.0.0',
 'orjson>=3.7.2,<4.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'typer>=0.4.1,<0.5.0']

entry_points = \
{'console_scripts': ['openapi-python-generator = '
                     'openapi_python_generator.__main__:main']}

setup_kwargs = {
    'name': 'openapi-python-generator',
    'version': '0.2.1.dev1656066202',
    'description': 'Openapi Python Generator',
    'long_description': "# Openapi Python Generator\n\n[![PyPI](https://img.shields.io/pypi/v/openapi-python-generator.svg)][pypi_]\n[![Status](https://img.shields.io/pypi/status/openapi-python-generator.svg)][status]\n[![Python Version](https://img.shields.io/pypi/pyversions/openapi-python-generator)][python version]\n[![License](https://img.shields.io/pypi/l/openapi-python-generator)][license]\n\n[![Read the documentation at https://openapi-python-generator.readthedocs.io/](https://img.shields.io/readthedocs/openapi-python-generator/latest.svg?label=Read%20the%20Docs)][read the docs]\n[![Tests](https://github.com/MarcoMuellner/openapi-python-generator/workflows/Tests/badge.svg)][tests]\n[![Codecov](https://codecov.io/gh/MarcoMuellner/openapi-python-generator/branch/main/graph/badge.svg)][codecov]\n\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)][pre-commit]\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)][black]\n\n[pypi_]: https://pypi.org/project/openapi-python-generator/\n[status]: https://pypi.org/project/openapi-python-generator/\n[python version]: https://pypi.org/project/openapi-python-generator\n[read the docs]: https://openapi-python-generator.readthedocs.io/\n[tests]: https://github.com/MarcoMuellner/openapi-python-generator/actions?workflow=Tests\n[codecov]: https://app.codecov.io/gh/MarcoMuellner/openapi-python-generator\n[pre-commit]: https://github.com/pre-commit/pre-commit\n[black]: https://github.com/psf/black\n\n## Features\n\n- Easy code generation for OpenAPI 3.0.0+ APIs\n- Async and Sync code generation support (with the help of [httpx](https://pypi.org/project/httpx/))])\n- Typed services and models for your convinience\n- Support for HttpBearer authentication\n- Python only\n- Usage as CLI tool or as a library\n\n## Requirements\n\n- Python 3.7+\n\n## Installation\n\nYou can install _Openapi Python Generator_ via [pip] from [PyPI]:\n\n```console\n$ pip install openapi-python-generator\n```\n\n## Usage\n\nPlease see the [Command-line Reference] for details.\n\n## Roadmap\n\n- Support for all commonly used http libraries in the python ecosystem (requests, urllib, ...)\n- Support for multiple languages\n- Support for multiple authentication schemes\n\n## Contributing\n\nContributions are very welcome.\nTo learn more, see the [Contributor Guide].\n\n## License\n\nDistributed under the terms of the [MIT license][license],\n_Openapi Python Generator_ is free and open source software.\n\n## Issues\n\nIf you encounter any problems,\nplease [file an issue] along with a detailed description.\n\n## Credits\n\nThis project was generated from [@cjolowicz]'s [Hypermodern Python Cookiecutter] template.\n\n[@cjolowicz]: https://github.com/cjolowicz\n[pypi]: https://pypi.org/\n[hypermodern python cookiecutter]: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n[file an issue]: https://github.com/MarcoMuellner/openapi-python-generator/issues\n[pip]: https://pip.pypa.io/\n\n<!-- github-only -->\n\n[license]: https://github.com/MarcoMuellner/openapi-python-generator/blob/main/LICENSE\n[contributor guide]: https://github.com/MarcoMuellner/openapi-python-generator/blob/main/CONTRIBUTING.md\n[command-line reference]: https://openapi-python-generator.readthedocs.io/en/latest/usage.html\n",
    'author': 'Marco Müllner',
    'author_email': 'muellnermarco@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MarcoMuellner/openapi-python-generator',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
